use core::fmt;
use std::{process::exit, time::Instant};

//use steel::steel_vm::engine::Engine;
use anyhow::{Context, Error, Result};
use lisp::{LispVMData, OpCode, Script};
use log::{debug, error, info};
use rodio::{
    cpal::{self, traits::HostTrait}, Device, DeviceTrait, OutputStream, OutputStreamHandle
};
use slint::{Brush, Color, Model, ModelRc, VecModel};

pub mod lisp;

slint::include_modules!();

trait UnwrapToLog<T, E> {
    fn unwrap_to_log(self) -> T
    where
        E: fmt::Debug;
}

impl<T, E> UnwrapToLog<T, E> for Result<T, E> {
    fn unwrap_to_log(self) -> T
    where
        E: fmt::Debug,
    {
        match self {
            Ok(x) => x,
            Err(e) => {
                error!("{e:?}");
                exit(1);
            }
        }
    }
}

trait RustMainWindowFunctions {
    fn connect(&self, node1: i32, port1: i32, node2: i32, port2: i32, brush: Brush, toggle: bool);
    fn add_connection(&self, connection: Connection, toggle: bool);
    fn add_node(&self, node: Node);
}

impl RustMainWindowFunctions for MainWindow {
    fn connect(&self, node1: i32, port1: i32, node2: i32, port2: i32, brush: Brush, toggle: bool) {
        self.add_connection(
            Connection {
                node1,
                port1,
                node2,
                port2,
                color: brush,
            },
            toggle,
        );
    }

    fn add_connection(&self, connection: Connection, toggle: bool) {
        let default = &VecModel::default();
        let connections = self.get_connections();
        let connections_ref = connections.as_any().downcast_ref().unwrap_or(default);
        let mut existing_connections: Vec<Connection> = connections_ref.iter().collect();

        let duplicate_index = existing_connections.iter().position(|r: &Connection| {
            r.node1 == connection.node1
                && r.node2 == connection.node2
                && r.port1 == connection.port1
                && r.port2 == connection.port2
        });

        match duplicate_index {
            Some(x) => {
                if toggle {
                    existing_connections.remove(x);
                }
            }
            None => {
                existing_connections.push(connection);
            }
        }

        info!("{existing_connections:?}");

        self.set_connections(ModelRc::new(VecModel::from(existing_connections)));
    }

    fn add_node(&self, node: Node) {
        let default = &VecModel::default();
        let nodes = self.get_nodes();
        let nodes_ref = nodes.as_any().downcast_ref().unwrap_or(default);
        let mut existing_nodes: Vec<Node> = nodes_ref.iter().collect();

        existing_nodes.push(node);

        self.set_nodes(ModelRc::new(VecModel::from(existing_nodes)));
    }
}

fn open_window_and_run_event_loop(device: Device, stream: OutputStreamHandle) -> Result<(), Error> {
    let main_window = MainWindow::new().with_context(|| "failed to create main window")?;

    main_window.set_window_title(format!("{} {}", env!("CARGO_PKG_NAME"), env!("CARGO_PKG_VERSION")).into());   

    let patchwire_start_time = Instant::now();

    main_window.add_node(Node {
        text: "Sine Wave".into(),
        // TODO: Node placement algorithm
        x: 0.0,
        y: 50.0,
        out_ports: ModelRc::new(VecModel::from(vec![
            Port {
                name: "a".into(),
                ..Default::default()
            },
        ])),
        ..Default::default()
    });

    main_window.add_node(Node {
        text: format!("Audio device {}", device.name().unwrap_or("default".to_owned())).into(),
        x: 200.0,
        y: 50.0,
        in_ports: ModelRc::new(VecModel::from(vec![
            Port {
                name: "L".into(),
                ..Default::default()
            },
            Port {
                name: "R".into(),
                ..Default::default()
            },
        ])),
        ..Default::default()
    });

    main_window.connect(0, 0, 1, 0, Brush::SolidColor(Color::from_rgb_u8(255, 255, 0)), true);
    main_window.connect(0, 0, 1, 1, Brush::SolidColor(Color::from_rgb_u8(255, 255, 0)), true);

    let mwr = main_window.as_weak();

    main_window.on_add_connection_helper(move |c: Connection| mwr.unwrap().add_connection(c, true));

    main_window
        .run()
        .with_context(|| "error in main window event loop")
}

// TODO: remove
fn script_test() {
    let script = Script::from_raw_bytecode(vec![
        OpCode::Input(LispVMData::Sample(1)),
        OpCode::Sub,
        //OpCode::JumpIfZero(4),
        OpCode::JumpIfEq(4, LispVMData::Sample(0)),
        OpCode::Jump(0),
        OpCode::Brk,
    ]);
    let input = vec![
        OpCode::Input(LispVMData::Sample(69)),
    ];

    let now = Instant::now();

    //
    let result = script.process(input, now);
    //

    let elapsed = now.elapsed();

    info!("Result: {:?}, Time: {:?}", result.unwrap_to_log(), elapsed);

}

fn main() {
    env_logger::init();
    //let mut steel_engine = Engine::new();
    // TODO: other config reading and setup goes here

    // TODO: allow user to select device
    debug!("Opening audio output stream");
    let default_device = cpal::default_host()
        .default_output_device()
        .with_context(|| "failed to find default audio device")
        .unwrap_to_log();

    info!("Found audio device with name: \"{}\"", default_device.name().unwrap_or("default".to_string()));

    let (_, stream) = OutputStream::try_from_device(&default_device)
        .with_context(|| "failed to open default audio device")
        .unwrap_to_log();

    script_test();

    debug!("Opening main window");
    open_window_and_run_event_loop(default_device, stream).unwrap_to_log();
}
